/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HTMLFormatter.java,v 1.8 2006/12/22 19:39:46 jkjome Exp $
 */

package org.enhydra.xml.io;

import java.io.IOException;

import org.enhydra.xml.dom.DOMAccess;
import org.enhydra.xml.dom.DOMOps;
import org.w3c.dom.Attr;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.Text;

/*
 * FIXME:
 *  - Need to check for optional attributes.
 *  - Give control over newline in OutputOptions or maybe use a PrintWriter???
 *  - Need option to include HTML 4.0 entity references.
 *  - script and style child special handling.
 *  - Need to output HTML header.
 */


/**
 * Formatter for outputting a HTML DOM as a HTML text document.
 */
final class HTMLFormatter extends BaseDOMFormatter implements Formatter {
    /**
     * Default XML encoding.
     */
    private static final String DEFAULT_XML_ENCODING = "ISO-8859-1";

    /**
     * Table use to optimized checking for characters that should be
     * represented as entity references.
     */
    private static final boolean[] ENTITY_QUICK_CHECK
        = new boolean[MAX_ENTITY_QUICK_CHECK_CHAR+1];

    /**
     * Should SPAN ID attributes be dropped?
     */
    private final boolean fDropSpanIds;

    /**
     * Flag that indicates the ID attribute should be dropped for the current
     * element.
     */
    private boolean fDropThisId;
    
    /**
     * Flag that indicates wether to use all named entites for all HTML 4.0 character
     * entities or not.
     */
    private boolean fUseHTML4Entities;

    /**
     * Nesting count for elements that don't have their content formatted.
     * This is done for pre, script, and style elements.  The contents of these
     * elements are output as-is.
     */
    private int fNoFormatNestCount;

    /**
     * Nesting count for script and style elements.  This goes one step beyond
     * fNoFormatNestCount, in that it allows us to avoid resolving special
     * characters or entities, which may exist inside script or style elements,
     * but be meant literally.  Essentially, it's a CDATA Section without being
     * marked as such officially.
     */
    private int fNoFormatScriptStyleNestCount;

    /**
     * Indicates that a text node has just been handled
     */
    private boolean fHandledText;

    /**
     * Indicates that the next sibling is a Text node
     */
    private boolean fNextSiblingText;

    /**
     * Static constructor.
     */
    static {
        for (char ch = 0; ch <= MAX_ENTITY_QUICK_CHECK_CHAR; ch++) {
            ENTITY_QUICK_CHECK[ch] = (HTMLEntities.charToEntity(ch) != null);
        }
    }

    /**
     * Constructor.
     */
    public HTMLFormatter(Node node,
                         OutputOptions outputOptions,
                         boolean forPreFormatting) {
        super(node, outputOptions, forPreFormatting, DEFAULT_XML_ENCODING, ENTITY_QUICK_CHECK);
        fDropSpanIds = fOptions.getDropHtmlSpanIds();
        fUseHTML4Entities = fOptions.getUseHTML4Entities();
    }

    /**
     * Get the default OutputOptions for a document formatter with this
     * formatter.  The encoding will not be set, which signals to use the
     * default encoding.
     */
    static OutputOptions getDefaultOutputOptions() {
        return new OutputOptions();  // Nothing special
    }

    /**
     * @see BaseDOMFormatter#getCharacterEntity
     */
    protected final String getCharacterEntity(char textChar) {
        if (fUseHTML4Entities) { 
            return HTMLEntities.charToEntity4(textChar);
        }
        return HTMLEntities.charToEntity(textChar);
    }

    /**
     * Determine if an attribute's value should be printer. Those that don't
     * normally have values only get them if one was explictly supplied.
     */
    private boolean printableAttrValue(Attr attr) {
        return (!(HTMLElements.isBooleanAttr(attr.getName())));
    }

    /**
     * Handler called for Document nodes.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleDocument
     */
    public void handleDocument(Document document) throws IOException {
        fTraverser.processDocumentType(document);
        fTraverser.processChildren(document);
    }

    /**
     * Handler called for Document nodes; writes the DOCTYPE specification.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleDocumentType
     */
    public void handleDocumentType(DocumentType documentType) throws IOException {
        if (fOptions.getOmitDocType()) {
            return; // Don't output DOCTYPE.
        }

        if ((fPublicId == null) && (fSystemId == null)) {
            return; // No DOCTYPE
        }
        //Must get element with out expanding if a LazyDOM.
        Element docElement =  DOMAccess.accessDocumentElement(fDocument);
        if (docElement == null) {
            throw new XMLIOError("Document has DocumentType, with out having a root element"); 
        }

        String tagName = docElement.getTagName();

        fOut.write("<!DOCTYPE ");
        fOut.write(tagName);
        
        if (fPublicId != null) {
            fOut.write(" PUBLIC \"");
            fOut.write(fPublicId);
            fOut.write("\"");
        }
        if (fSystemId != null) {
            if (fPublicId == null) {
                fOut.write(" SYSTEM");
            }
            fOut.write(" \"");
            fOut.write(fSystemId);
            fOut.write("\"");
        }

        fOut.write('>');
        writeln();
    }

    /**
     * Handler called for DocumentFragment nodes; just process children
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleDocumentFragment
     */
    public void handleDocumentFragment(DocumentFragment documentFragment) {
        fTraverser.processChildren(documentFragment);
    }

    /**
     * Handler called for Attr nodes.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleAttr
     */
    public void handleAttr(Attr attr) throws IOException {
        String name = attr.getName();
        if (!(fDropThisId && name.equals("id"))) {
            fOut.write(' ');
            fOut.write(name);
            if (printableAttrValue(attr)) {
                writeAttributeValue(attr);
            }
        }
    }

    /**
     * Write an element open tag.
     */
    protected final void writeOpenTag(Element element, 
                                      String tagName,
                                      boolean hasChildren) throws IOException {
        String formattedTag = null;
        if (fPrettyPrinting) {
            //Revisit - excluding this newline when formatting HTML
            //and preserve space is true because of differences, presumably, in
            //the NekoHTML parser where seems to store extra newlines between
            //elements.  Until a better solution is found, this will have to
            //suffice.
            if (!fPreserveSpace && fNextSiblingText && fNoFormatNestCount == 0) {
                fOut.write('\n');
            }
            fNextSiblingText = !isNotTextOrIsTextButAllWhitespace(DOMAccess.accessNextSibling(fDocument, element));
        }

        if (fUsePreFormattedElements && (element instanceof PreFormattedText)) {
            formattedTag = ((PreFormattedText)element).getPreFormattedText();
        }
        if (formattedTag != null) {
            fOut.write(formattedTag);
            fPreFormattedElementCount++;
        } else {
            if (fPrettyPrinting && isNotTextOrIsTextButAllWhitespace(element.getPreviousSibling())) {
                printIndent();
            }

            fDropThisId = fDropSpanIds && tagName.equals("SPAN");
            fOut.write('<');
            fOut.write(tagName);
            fTraverser.processAttributes(element);
            if (!hasChildren) {
                fHandledText = true; //spoof to avoid unnecessary indent later
            }
            fOut.write('>');
            fDynamicFormattedElementCount++;
            if (fPrettyPrinting && hasChildren && isNotTextOrIsTextButAllWhitespace(element.getFirstChild())) {
                fOut.write('\n');
            }
        }
    }

    /**
     * Note: the node type check and whitespace check are only performed if the
     * node is not within a parent having xml:space="preserve".  If this
     * condition is satisfied and the type check determines the node is a Text
     * node, the whitespace check is only performed if
     * OutputOptions.getPreserveSpace() is 'false'.
     * 
     * @param node the node to test
     * @return true if not within a parent having xml:space attribute with value
     *         'preserve' and the node is not Text or is Text but all whitespace,
     *         otherwise false
     */
    private boolean isNotTextOrIsTextButAllWhitespace(Node node) {
        if (fNoFormatNestCount > 0) return false;
        boolean isTextNode = DOMOps.isNodeOfType(node, Node.TEXT_NODE);
        if (!isTextNode || (isTextNode && !fPreserveSpace && !isTextPreformatted(node) && ((Text)node).getData().trim().length() == 0)) {
            return true;
        }
        return false;
    }

    /**
     * No check is made for the node being a Text node, but this method should
     * only be used in such cases as it is a Text node.  Didn't want to
     * implement the extra check for performance reasons since, in most cases,
     * the check for node type Text would have already been performed. 
     * 
     * @param node the node to check for instanceof PreFormattedText if
     *             respective boolean flags are true
     * @return true if respective flags to use preformatted text are true and
     *         the node is, in fact, an instanceof PreFormattedText
     */
    private boolean isTextPreformatted(Node node) {
        return ((fProcessingAttr && fUsePreFormattedAttrText) || fUsePreFormattedText) && node instanceof PreFormattedText;
    }

    /**
     * Handler called for Element nodes.
     * 
     * <p>
     * This optionally corrects problem cases for browsers:
     * </p>
     * <ul>
     * <li>
     * ID attributes are dropped from SPAN tags.  This cause Internet Explorer
     * 4.0 to get confused on keep-alive connections.
     * </li>
     * </ul>
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleElement
     */
    public void handleElement(Element element) throws IOException {
        String tagName = element.getTagName();
        boolean hasChildren = element.hasChildNodes();
        boolean isPreserveSpaceTag = false;
        boolean isScriptStyleTag = false;

        // Start Barracuda Kludge ======
        // check the element to see if it contains a
        // attribute "visdom".
        // (org.enhydra.barracuda.core.comp.BComponent.VISIBILITY_MARKER)
        // This controls DOM visibility. If this value exists and does not
        // match to true, don't print this particular node.
        // Note: This should be made generic, but for now...
        //TODO - consider getting rid of this, I don't think it's used much or at all in Barracuda
        Attr attr = DOMAccess.accessAttribute(fDocument, element, null, "visdom");
        if (attr != null && !(Boolean.valueOf(DOMAccess.accessAttributeValue(fDocument, attr)).booleanValue())) return;        
        // End Barracuda Kludge ======

        if (hasChildren) {
            isPreserveSpaceTag = HTMLElements.isPreserveSpaceTag(tagName);
            if (isPreserveSpaceTag) {
                fNoFormatNestCount++;
                isScriptStyleTag = HTMLElements.isScriptStyleTag(tagName);
                if (isScriptStyleTag) fNoFormatScriptStyleNestCount++;
            }
        }

        writeOpenTag(element, tagName, hasChildren);

        // Process children and close.
        boolean processedChildren = false;
        boolean processedLastChildElement = false;
        if (hasChildren) {
            fTraverser.processChildren(element);
            processedChildren = true;
            //below only matters when using pretty printing AND preserving space (nothing to do with xml:space).  If preserve space is false, avoid the hit of the nextSibling check (checking if the current element is a last child)
            if (fPrettyPrinting && fPreserveSpace && DOMAccess.accessNextSibling(fDocument, element) == null) {
                processedLastChildElement = true;
            }
        }

        if (hasChildren && isPreserveSpaceTag) {
            fNoFormatNestCount--;
            if (isScriptStyleTag) fNoFormatScriptStyleNestCount--;
        }

        if (fHandledText) {
            fHandledText = false;
        } else {
            printIndent();
        }

        if (processedChildren || HTMLElements.hasCloseTag(tagName)) {
            fOut.write("</");
            fOut.write(tagName);
            fOut.write('>');
        }

        if (fPrettyPrinting && (!fNextSiblingText || processedLastChildElement)) {
            fOut.write('\n');
        }
    }

    /**
     * Handler called for ProcessingInstruction nodes.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleProcessingInstruction
     */
    public void handleProcessingInstruction(ProcessingInstruction pi) throws IOException {
        throw new XMLIOError("Unexpected call to handleProcessingInstruction");
    }

    /**
     * Handler called for CDATASection nodes.
     * <p>Non-standard extension: outputs data as-is.</p>
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleCDATASection
     */
    public void handleCDATASection(CDATASection cdata) throws IOException {
        fOut.write(cdata.getData());
    }

    /**
     * Handler called for Text nodes.
     * 
     * @see org.enhydra.xml.dom.DOMTraversal.Handler#handleText
     */
    public final void handleText(Text text) throws IOException {
        if (fNoFormatScriptStyleNestCount > 0) {
            fOut.write(text.getData());
            fHandledText = true;
            return;
        } else {
            if (!fPreserveSpace && fNoFormatNestCount == 0 && !isTextPreformatted(text)) {
                if (text.getData().trim().length() == 0) {
                    if (!fPrettyPrinting) {
                        fOut.write(' '); //need to provide at least a space if not using pretty printing (which would provide the necessary space via line breaks)
                    }
                    //intentionally not setting "fHandledText = true"
                    return;
                }
            }
        }
        super.handleText(text);
        fHandledText = true;
    }

}
